<?php

namespace InSegment\WorkersFramework\Services;

use Shavshukov\EventAppBase\ApplicationInterface;
use Shavshukov\EventAppBase\Interfaces\ModelInterface;
use React\Promise;

use InSegment\WorkersFramework\Interfaces\ConnectionListInterface;
use InSegment\WorkersFramework\Interfaces\QueueMessageInterface;
use InSegment\WorkersFramework\Interfaces\WorkerListInterface;
use InSegment\WorkersFramework\Interfaces\QueueInterface;
use InSegment\WorkersFramework\Interfaces\JobConfigInterface;
use InSegment\WorkersFramework\Services\MainConsumer;
use InSegment\WorkersFramework\Http\Server;
use InSegment\WorkersFramework\Models\State;
use InSegment\WorkersFramework\Queues\Worker;
use InSegment\WorkersFramework\Queues\LoadMeasure;
use InSegment\WorkersFramework\Queues\ServiceMessenger;

class Manager implements WorkerListInterface
{
    /**
     * Application instance
     *
     * @var \Shavshukov\EventAppBase\ApplicationInterface
     */
    protected $app;
    
    /**
     * Instance of queue driver connected to the channel
     *
     * @var \InSegment\WorkersFramework\Interfaces\ConnectionListInterface
     */
    protected $connections;
    
    /**
     * HTTP server instance
     *
     * @var \InSegment\WorkersFramework\Http\Server
     */
    protected $httpServer;
    
    /**
     * Configuration of each worker
     *
     * @var array
     */
    protected $workersConfig;
    
    /**
     * Consumer for main queue
     *
     * @var \InSegment\WorkersFramework\Services\MainConsumer
     */
    protected $mainConsumer;
    
    /**
     * Consumer for state queue
     *
     * @var \InSegment\WorkersFramework\Services\StateConsumer
     */
    protected $stateConsumer;
    
    /**
     * State instance
     *
     * @var \InSegment\WorkersFramework\Models\State
     */
    protected $state;
    
    /**
     * All workers initialised
     *
     * @var \InSegment\WorkersFramework\Queues\Worker[]
     */
    protected $workers;
    
    /**
     * Worker processes
     *
     * @var \InSegment\WorkersFramework\Services\WorkerProcess[]
     */
    protected $workerProcesses;
    
    /**
     * Messengers for workers service processes
     *
     * @var \InSegment\WorkersFramework\Queues\ServiceMessenger[]
     */
    protected $workerMessengers;
    
    /**
     * Load measure instance
     *
     * @var \InSegment\WorkersFramework\Queues\LoadMeasure
     */
    protected $loadMeasure;
    
    /**
     * Messengers to the other nodes of multi-master configuration
     *
     * @var \InSegment\WorkersFramework\Queues\ServiceMessenger[]
     */
    protected $otherNodesMessengers;
    
    /**
     * Constructor
     * 
     * @param \Shavshukov\EventAppBase\ApplicationInterface $app
     * @param \InSegment\WorkersFramework\Interfaces\ConnectionListInterface $connections
     * @throws \Exception
     */
    public function __construct(ApplicationInterface $app, ConnectionListInterface $connections)
    {
        $this->app = $app;
        $this->connections = $connections;
        $this->httpServer = new Server($app);
        $this->workersConfig = $app->config('workers', []);
        $statePath = implode(DIRECTORY_SEPARATOR, [$this->app->getAppPath(), $app->config('state_file', 'storage/state.yaml')]);
        $this->state = new State($this->app->getFileReader(), $statePath);
        $this->loadMeasure = new LoadMeasure(array_keys($this->workersConfig));
        $this->mainConsumer = new MainConsumer(
            $connections->getQueue($app->config('main_queue')),
            new Distributor($connections, $this),
            $app->config()->getJobList()
        );
        $this->stateConsumer = new StateConsumer(
            $connections->getQueue($app->config('service_queue')),
            $this->state
        );
        $this->initNodes();
        $this->initEvents();
        $this->initWorkers();
        $app->service()->bind('manager', $this);
    }

    /**
     * Init worker consumers and Http server
     * 
     * @return \React\Promise\PromiseInterface
     */
    public function manage(): Promise\PromiseInterface
    {
        $promises = [];
        $this->state->load();
        
        // notify nodes about my current state - so when all the nodes are restarted, the state of all will be the most recent
        if ($this->state->has('timestamp')) {
            $this->sendToNodes($this->state->all(), false);
        }

        $promises[] = $this->stateConsumer->activate();
        $promises[] = $this->httpServer->serve();
        
        if (!$this->state->get('locked')) {
            $promises[] = $this->activateConsuming(false);
        }
        
        return Promise\all($promises);
    }

    /**
     * Activate consuming of the main queue
     * 
     * @param bool $setState
     * @return \React\Promise\PromiseInterface
     */
    public function activateConsuming(bool $setState = true): Promise\PromiseInterface
    {
        if ($setState && $this->state->get('locked')) {
            $modification = ['locked' => false];
            $this->state->modify($modification)->save();
            $this->sendToNodes($modification);
        }
        
        return $this->mainConsumer->activate();
    }
    
    /**
     * Activate particular worker
     * 
     * @param string $workerName
     * @param bool $setState
     * @return \React\Promise\PromiseInterface
     * @throws \Exception
     */
    public function activateWorker(string $workerName, bool $setState = true): Promise\PromiseInterface
    {
        if (!isset($this->workers[$workerName])) {
            throw new \Exception("Worker {$workerName} does not exists");
        }
        
        $worker = $this->workers[$workerName];
        $this->loadMeasure->add($workerName);
        
        if ($setState && $this->state->get("workers.{$workerName}.locked")) {
            $modification = [
                'workers' => [
                    $workerName => [
                        'locked' => false
                    ]
                ]
            ];
            $this->state->modify($modification)->save();
            $this->sendToNodes($modification);
        }
        
        if ($worker->isPresentOnNode() && !($this->state->get('locked') || $worker->isActive())) {
            $this->workerProcesses[$workerName]->start();
        }
        
        return Promise\resolve();
    }

    /**
     * Deactivate consuming of the main queue
     * 
     * @param bool $setState
     * @return \React\Promise\PromiseInterface
     */
    public function deactivateConsuming(bool $setState = true): Promise\PromiseInterface
    {
        echo "Deactivating consume of the main queue...\n";
        
        if ($setState && !$this->state->get('locked')) {
            $modification = ['locked' => true];

            $this->state->modify($modification)->save();
            $this->sendToNodes($modification);
        }
        
        foreach ($this->workers as $worker) {
            if ($worker->isActive()) {
                $worker->modifyStatus(['workerState' => 'stopping...']);
            }
        }
        
        return $this->mainConsumer->deactivate();
    }
    
    /**
     * Deactivate particular worker
     * 
     * @param string $workerName
     * @param bool $setState
     * @return \React\Promise\PromiseInterface
     * @throws \Exception
     */
    public function deactivateWorker(string $workerName, bool $setState = true): Promise\PromiseInterface
    {
        if (!isset($this->workers[$workerName])) {
            throw new \Exception("Worker {$workerName} does not exists");
        }
        
        $worker = $this->workers[$workerName];
        $this->loadMeasure->remove($workerName);
        
        $modification = [
            'workers' => [
                $workerName => [
                    'locked' => true
                ]
            ]
        ];
        
        if ($setState && !$this->state->get("workers.{$workerName}.locked")) {
            $this->state->modify($modification)->save();
            $this->sendToNodes($modification);
        }
        
        if (!$this->state->get('locked') && $worker->isActive()) {
            $worker->modifyStatus(['workerState' => 'stopping...']);
            echo "Sending shutdown update to worker {$workerName}...\n";
            return $this->workerMessengers[$workerName]->sendUpdates($modification);
        }
        
        return Promise\resolve();
    }
    
    /**
     * Get queue status as array
     * 
     * @return array
     */
    public function getStatus(): array
    {
        $workerStatus = [];
        foreach ($this->workers as $worker) {
            $workerStatus[$worker->getName()] = $worker->getStatus();
        }

        return [
            'workers_config' => $this->workersConfig,
            'consumers' => $workerStatus,
            'state' => $this->state->all(),
        ];
    }

    /**
     * {@inheritDoc}
     */
    public function getWorkers(array $filter = null): array
    {
        $availableWorkers = [];
        foreach ($this->workers as $worker) {
            $workerName = $worker->getName();
            if (!$this->state->isWorkerLocked($workerName) && ($filter === null || in_array($workerName, $filter))) {
                $availableWorkers[] = $worker;
            }
        }
        
        return $availableWorkers;
    }

    /**
     * {@inheritDoc}
     */
    public function getLoadMeasure(): LoadMeasure
    {
        return $this->loadMeasure;
    }
    
    /**
     * Send state updates to other nodes of the cluster
     * 
     * @param array $modification
     * @param bool $updateTimestamp
     * @return \React\Promise\PromiseInterface
     */
    protected function sendToNodes(array $modification, bool $updateTimestamp = true): Promise\PromiseInterface
    {
        if (count($this->otherNodesMessengers)) {
            echo "Sending state updates to other nodes...\n";
            $promises = [];
            foreach ($this->otherNodesMessengers as $messenger) {
                $promises[] = $messenger->sendUpdates($modification, $updateTimestamp);
            }

            return Promise\all($promises)->then(function() {
                echo "Updates to other nodes were sent successfully...\n";
            });
        }
        
        return Promise\resolve();
    }
    
    /**
     * Prepare all workers and their processes
     * 
     * @return null
     */
    protected function initWorkers()
    {
        $workers = [];
        $workerProcesses = [];
        $workerMessengers = [];
        
        $nodeName = $this->app->config('node_name');
        foreach ($this->workersConfig as $workerName => $config) {
            $worker = new Worker($workerName, $config, $nodeName);
            $workerProcesses[$workerName] = new WorkerProcess($this->app, $this->state, $worker);
            $workerMessengers[$workerName] = new ServiceMessenger($this->connections->getQueue($worker->getServiceQueueName()));
            $workers[$workerName] = $worker;
        }
        
        $this->workers = $workers;
        $this->workerProcesses = $workerProcesses;
        $this->workerMessengers = $workerMessengers;
    }
    
    /**
     * Init other nodes for multi-master configuration
     * 
     * @return null
     */
    protected function initNodes()
    {
        $otherNodesMessengers = [];
        $nodeName = $this->app->config('node_name');
        $multiMaster = $this->app->config('multi_master');
        
        if ($multiMaster) {
            foreach ($multiMaster as $otherNodeName => $otherNodeConfig) {
                $otherNodeServiceQueue = $otherNodeConfig['service_queue'];
                
                if ($otherNodeName !== $nodeName) {
                    $otherNodesMessengers[$otherNodeName] = new ServiceMessenger($this->connections->getQueue($otherNodeServiceQueue));
                }
            }
        }
        
        $this->otherNodesMessengers = $otherNodesMessengers;
    }
    
    /**
     * Init events to listen on the consumers
     * 
     * @return null
     */
    protected function initEvents()
    {
        $this->mainConsumer->on('starting', function(QueueInterface $queue) {
            echo "Starting main consumer on queue '{$queue->getQueueName()}'...\n";
        });
        
        $this->mainConsumer->on('started', function(QueueInterface $queue, string $consumerTag) {
            echo "Started main consumer with tag '{$consumerTag}' on queue '{$queue->getQueueName()}'\n";
            $this->startEnabledWorkers();
        });
        
        $this->mainConsumer->on('publishing', function (Worker $worker, JobConfigInterface $job, QueueMessageInterface $newMessage) {
            echo "Distributing message with job {$job->getName()} for worker {$worker->getName()} into queue {$newMessage->getQueueName()}\n";
        });
        
        $this->mainConsumer->on('published', function (Worker $worker, JobConfigInterface $job, QueueMessageInterface $newMessage) {
            $this->loadMeasure->increment($worker->getName(), max([$job->getWeight(), 1]));
        });
        
        $this->mainConsumer->on('stopping', function(QueueInterface $queue, string $consumerTag) {
            echo "Stopping main consumer with tag '{$consumerTag}' on queue '{$queue->getQueueName()}'...\n";
        });
        
        $this->mainConsumer->on('stopped', function(QueueInterface $queue, string $consumerTag) {
            echo "Stopped main consumer with tag '{$consumerTag}' on queue '{$queue->getQueueName()}'\n";
            $this->stopEnabledWorkers(['locked' => true]);
        });
        
        $this->stateConsumer->on('starting', function(QueueInterface $queue) {
            echo "Starting state consumer on queue: {$queue->getQueueName()}\n";
        });
        
        $this->stateConsumer->on('started', function(QueueInterface $queue, string $consumerTag) {
            echo "Started state consumer on queue: {$queue->getQueueName()}, tag: {$consumerTag}\n";
        });
        
        $this->stateConsumer->on('updates', function (ModelInterface $updates) {
            echo "Received external state update\n";
            $this->processNodeStateUpdates($updates);
        });
    }
    
    /**
     * Process external updates for the state of the current node
     * 
     * @param \Shavshukov\EventAppBase\Interfaces\ModelInterface $updates
     */
    protected function processNodeStateUpdates(ModelInterface $updates)
    {
        $consumingActivate = false;
        $consumingDeactivate = false;
        $lockWorkers = [];
        $unlockWorkers = [];
        $save = false;

        if ($updates->has('locked')) {
            $myLocked = $this->state->get('locked');
            $updatedLocked = $updates->get('locked');

            if ($updatedLocked !== null && $myLocked !== $updatedLocked) {
                if ($updatedLocked) {
                    echo "To lock all\n";
                    $consumingDeactivate = true;
                } else {
                    echo "To unlock all\n";
                    $consumingActivate = true;
                }
            }
            
            $save = true;
        }

        if ($updates->has('workers')) {
            $updatedWorkers = (array) $updates->get('workers');
            foreach ($updatedWorkers as $workerName => $workerConfigUpdate) {
                if (!isset($this->workers[$workerName])) {
                    continue;
                }
                
                $myLockedWorker = $this->state->get("workers.{$workerName}.locked");
                $updatedLockedWorker = $workerConfigUpdate['locked'] ?? null;

                if ($updatedLockedWorker !== null && $myLockedWorker !== $updatedLockedWorker) {
                    if ($updatedLockedWorker) {
                        echo "To lock {$workerName}\n";
                        $lockWorkers[] = $workerName;
                    } else {
                        echo "To unlock {$workerName}\n";
                        $unlockWorkers[] = $workerName;
                    }
                }
            }
            
            $save = true;
        }

        $this->state->modify($updates->all());
        
        if ($save) {
            $this->state->save();
        }
        
        if ($consumingActivate) {
            $this->activateConsuming(false);
        } else if ($consumingDeactivate) {
            $this->deactivateConsuming(false);
        }
        
        foreach ($lockWorkers as $workerName) {
            $this->deactivateWorker($workerName, false);
        }
        
        foreach ($unlockWorkers as $workerName) {
            $this->activateWorker($workerName, false);
        }
    }
    
    /**
     * Start workers which are active
     * 
     * @return null
     */
    protected function startEnabledWorkers()
    {
        echo "Starting worker processes...\n";
        $enabledWorkers = $this->getWorkers();
        foreach ($enabledWorkers as $enabledWorker) {
            if ($enabledWorker->isPresentOnNode() && !$enabledWorker->isActive()) {
                $this->workerProcesses[$enabledWorker->getName()]->start();
            }
        }
    }
    
    /**
     * Stop workers which are active and sent them modification of the state
     * 
     * @param array $modification
     * @return \React\Promise\PromiseInterface
     */
    protected function stopEnabledWorkers(array $modification): Promise\PromiseInterface
    {
        echo "Sending shutdown updates to worker service queues...\n";
        $promises = [];
        foreach ($this->workers as $workerName => $worker) {
            if ($worker->isActive()) {
                $promises[] = $this->workerMessengers[$workerName]->sendUpdates($modification);
            }
        }

        return Promise\all($promises)->then(function() {
            echo "Updates to workers were sent successfully...\n";
        });
    }
    
}
